Object subclass: Solution [
    | gap aligned_x aligned_y col1 col2 col3|
    <comment:
        'A solution for string similarity by Hirschberg Algorithm'>
        
    Solution class >> new [
        <category: 'instance creation'>
        | r |
        r := super new.
        ^r
    ]
    
    init: x [
        <category: 'initialization'>
        | length |
        gap := 7.           "gap penalty"
        aligned_x := ''.    "a string to save aligned x"
        aligned_y := ''.    "a string to save aligned y"
        
        "In DP part, we only use two columns. Advantage is the space compelexity is linear. "
        "Disadvantage is the history is not saved, so you can't get the aligned results by these two colums"
        length := x size + 1.
        col1 := Array new:(length).     "an array which is used for left part DP"
        col2 := Array new:(length).     "an array which is used for right part DP"
        col3 := Array new:(length).     "an array which is used for both left and right DP"
    ]
    
    sol: x and: y [
        <category: 'first initial then solve'>
        self init:x.                        "first we initial the variables"
        ^(self hirschberg: x and: y)        "use hirschberg to solve it"
    ]
    
    hirschberg: x and: y [
        "get two strings: x and y. Align these two strings. Return the cost."
        | cost lx ly mid subxl subxr subyl subyr min_pos min_cost tmp|
        cost := 0.
        lx := x size.
        ly := y size.
        
        (lx = 0)
            ifTrue: [   "This is one of the end situation of divide and conquer. x is empty, so fill x with '|'"
                1 to: ly do:[:i|
                    aligned_x := (aligned_x, '|').
                    aligned_y := (aligned_y, (y at: i) asString).
                    cost := cost + gap
                ]
            ]
            ifFalse: [
                (ly = 0)
                    ifTrue: [   "This is one of the end situation of divide and conquer. y is empty, so fill y with '|'"
                        1 to: lx do:[:i|
                            aligned_x := (aligned_x, (x at: i) asString).
                            aligned_y := (aligned_y, '|').
                            cost := cost + gap
                        ]
                    ]
                    ifFalse: [
                        (ly = 1)
                            ifTrue: [   "This is one of the end situation of divide and conquer. y has one char, don't need to divide, call one_row function to solve"
                                cost := cost + (self needleman_one_row: x and: y)
                            ]
                            ifFalse: [  "y has more than one char, divide y into left and right, solve them recursively"
                                mid := ly // 2.
                                
                                subyl := y copyFrom:1 to: mid.
                                self left_row_needleman: x and: subyl.  "use DP to solve x and y-left"
                                subyr := y copyFrom:(mid + 1) to: ly.
                                self right_row_needleman: x and: subyr. "use DP to solve x and y-right"
                                
                                "In the solutions of DP, find the best divide point in x which leads to the minmum cost"
                                min_pos := 1.
                                min_cost := (col1 at: 1) + (col2 at: (lx+1)).
                                
                                2 to: (lx+1) do:[:i|
                                    tmp := (col1 at: i) + (col2 at: (lx+2-i)).
                                    (tmp < min_cost)
                                        ifTrue:[
                                            min_pos := i.
                                            min_cost := tmp
                                        ]
                                ].
                                
                                "divide x into x-left and x-rigt"
                                subxl := x copyFrom:1 to: (min_pos-1).
                                subxr := x copyFrom:min_pos to: lx.
                                cost := cost + (self hirschberg: subxl and: subyl). "align x-left and y-left"
                                cost := cost + (self hirschberg: subxr and: subyr)  "align x-right and y-right"
                            ].
                    ].
            ].
            
        ^cost
    ]
    
    
    needleman_one_row: x and: y [
        "y has one char, x is not empty, align x and y"
        | lx ly m n t1 t2 t3|

        lx := x size.
        ly := y size.

        "DP with col1 and col3"
        col1 at: 1 put: 0.
        2 to: (lx + 1) do: [:i|
            col1 at: i put: ((col1 at: (i-1)) + gap)
        ].
        
        col3 at: 1 put: ((col1 at: 1) + gap).
        2 to: (lx + 1) do: [:j|
            t1 := (col1 at: (j-1)) + (self get_cost:(x at:(j-1)) and: (y at: 1)).
            t2 := (col1 at: j) + gap.
            t3 := (col3 at: (j-1)) + gap.
            col3 at: j put: ((t1 min: t2) min: t3)  
        ].
        
        "Because DP doesn't save history. Here we trace back to find the aligned results."
        m := lx + 1.    "if n is 0, m is the position of the algined char in x"
        n := 1.         "a tag. 0 means char in y aligned with a char in x. 1 means continus while loop. -1 means stop while loop"
        
        "get m and n"
        [n > 0] whileTrue: [
            ((col3 at: m) = ((col1 at: m) + gap))
                ifTrue: [
                    n := -1
                ]
                ifFalse: [
                    ((col3 at: m) = ((col1 at: (m-1)) + (self get_cost:(x at: (m-1)) and: (y at: 1))))
                        ifTrue: [
                            n := 0.
                            m := m-1
                        ]
                        ifFalse: [
                            ((col3 at: m) = ((col3 at: (m-1)) + gap))
                                ifTrue: [
                                    m := m - 1  
                                ].    
                        ].
                ].
        ].
        
        "update aligned_x and aligned_y by m and n"
        1 to: (m-1) do: [:i|
            aligned_x := (aligned_x, (x at: i) asString).
            aligned_y := (aligned_y, '|')
        ].
        
        aligned_y := (aligned_y, (y at: 1) asString).
        (n = 0)
            ifTrue: [
                aligned_x := (aligned_x, (x at: m) asString).
                m := m + 1
            ]
            ifFalse: [
                aligned_x := (aligned_x, '|')
            ].
            
        m to: lx do: [:i|
            aligned_x := (aligned_x, (x at: i) asString).
            aligned_y := (aligned_y, '|')
        ].
        
        "return the cost of this alignment"
        ^(col3 at: (lx + 1))    
    ]
    
    
    left_row_needleman: x and: y [
        "use DP to align x and y. DP result saves in col1"
        | lx ly t1 t2 t3 tmp|
        lx := x size.
        ly := y size.
        
        col1 at: 1 put: 0.
        2 to: (lx + 1) do: [:i|
            col1 at: i put: ((col1 at: (i-1)) + gap)
        ].
        
        1 to: ly do: [:i|
            col3 at: 1 put: ((col1 at: 1) + gap).
            2 to: (lx + 1) do: [:j|
                t1 := (col1 at: (j-1)) + (self get_cost:(x at:(j-1)) and: (y at: i)).   "DP, index from small to large"
                t2 := (col1 at: j) + gap.
                t3 := (col3 at: (j-1)) + gap.
                col3 at: j put: ((t1 min: t2) min: t3)  
            ].
            
            tmp := col1.
            col1 := col3.
            col3 := tmp
        ]
    ]
    
    right_row_needleman: x and: y [
        "use DP to align x and y. DP result saves in col2"
        | lx ly t1 t2 t3 tmp|
        lx := x size.
        ly := y size.
        
        col2 at: 1 put: 0.
        2 to: (lx + 1) do: [:i|
            col2 at: i put: ((col2 at: (i-1)) + gap)
        ].
        
        2 to: (ly+1) do: [:i|
            col3 at: 1 put: ((col2 at: 1) + gap).
            2 to: (lx + 1) do: [:j|
                t1 := (col2 at: (j-1)) + (self get_cost:(x at:(lx-j+2)) and: (y at: (ly-i+2))). "DP, index from large to small"
                t2 := (col2 at: j) + gap.
                t3 := (col3 at: (j-1)) + gap.
                col3 at: j put: ((t1 min: t2) min: t3)  
            ].
            
            tmp := col2.
            col2 := col3.
            col3 := tmp
        ]
    ]
    
    
    get_cost: x and: y [
        ^(((x asciiValue) - (y asciiValue)) abs)    
    ]
    



]

| s str1 str2 c |
s := Solution new.
str1 := 'CTACPG'.
str2 := 'TACATG'.

c := s sol: str1 and: str2.
c printNl.
